# -*- coding: utf-8 -*-
import os
import errno


# TODO: on cleaning up take this into consideration:
# http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python > 2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



# -*- coding: utf-8 -*-
import os

from . import bash
from . import conf
from . import plugins
from . import shell


abspath = os.path.abspath
pjoin = os.path.join
exists = os.path.exists


class PathCalculator(object):  # stupid name....
    def __init__(self, place):
        self.place = place

    def __call__(self, *path):
        absolute_here = abspath(self.place)  # always in the place where it was called
        return abspath(pjoin(absolute_here, *path))


class Penv(bash.BashMixin, conf.ConfigurationMixin):
    def get_plugins(self, root_new):
        config = self.get_config(root_new)
        return plugins.load(config.PLUGIN_PLACES, config)

    # init
    #
    # has nothing todo with the rest, but I don't know yet where to
    # put it. Supposed to be invoked on "$> penv init" command.
    def init_script(self, root_new):
        result = []
        for plugin in self.get_plugins(root_new):
            scripts = plugin.init(root_new)
            result.extend(scripts or [])
        return "\n".join(result)

    def init(self, place):
        # by default create only ".penv" directory and append:
        # "eval "$(penv scan)""
        root_new = PathCalculator(place)
        config = self.get_config(root_new)
        shell.mkdir_p(root_new(config.TRIGGER))
        return self.init_script(root_new)
    # end / init

    def deactivate_bash_script(self, root_new, root_old):
        result = []
        for plugin in self.get_plugins(root_new):
            scripts = plugin.on_deactivate(root_new, root_old)
            result.extend(scripts or [])
        return "\n".join(result)

    def activate_bash_script(self, root_new, root_old):
        result = []
        for plugin in self.get_plugins(root_new):
            scripts = plugin.on_activate(root_new, root_old)
            result.extend(scripts or [])
        return "\n".join(result)

    def setup(self, place):
        old_pwd = os.environ.get("OLDPWD")
        root_new = PathCalculator(place)
        root_old = PathCalculator(old_pwd)
        config = self.get_config(root_new)
        script_stream = bash.BashStream()

        script_stream.writeln('# Scanning: %s | OLDPWD=%s' % (place, old_pwd))
        location = root_new(config.TRIGGER)
        venv_exists = exists(location)
        skip_searching = config.SKIP_FILE_EXISTS

        auto_venv_is_active = os.environ.get("PENV_IS_ACTIVE")
        if auto_venv_is_active:
            script_stream.writeln('#     '
                'previous env found (%s), so deactivating...' % location)
            script_stream.writeln(self.deactivate_bash_script(root_new, root_old))
            script_stream.writeln(self.bash.unset("PENV_IS_ACTIVE"))
            del os.environ["PENV_IS_ACTIVE"]
            script_stream.writeln('#     ...deactivation done.')

        if venv_exists and not skip_searching:
            script_stream.writeln('#     '
                'new env found (%s) and it shouldn\'t be '
                'skipped, so generating activation scripts' % location)

            # Information on where I'll look for plugins
            script_stream.writeln('#     %s' % ('#' * 30))
            script_stream.writeln('#     Following places will be checked for plugins existance:')
            for place in config.PLUGIN_PLACES:
                script_stream.writeln('#        %s' % place)
            script_stream.writeln('#     %s' % ('#' * 30))

            script_stream.writeln(self.activate_bash_script(root_new, root_old))
            script_stream.writeln(self.bash.export("PENV_IS_ACTIVE", "True"))
            script_stream.writeln('#     generation script done.')

        if skip_searching:
            script_stream.writeln('# Skipping searching because of: %s' % config.SKIP_FILE_PATH)

        return venv_exists, skip_searching, script_stream

    # TODO: change it to iterative version
    def lookup(self, place, parent_stream=None):
        parent_stream = parent_stream or bash.BashStream()

        venv_exists, skip_searching, stream = self.setup(place)
        parent_stream.writeln(stream.getvalue())

        if place == "/":
            return venv_exists, place, parent_stream.getvalue()

        if not venv_exists and not skip_searching:
            return self.lookup(abspath(pjoin(place, "..")), parent_stream)

        return venv_exists, place, parent_stream.getvalue()



# -*- coding: utf-8 -*-
import os
import click
import subprocess

from datetime import datetime

from .base import Penv
from . import VERSION


abspath = os.path.abspath
startup_script_bash = """
export PENV_SESSION_ID="$(python -c 'import uuid; print(uuid.uuid4())')"
export PENV_LOCK_FILE="/tmp/.penv-lock-$PENV_SESSION_ID"

function cd () {
    if [ -f $PENV_LOCK_FILE ]
    then
        builtin cd "$@"
    else
        touch $PENV_LOCK_FILE
        builtin cd "$@" && eval "$(penv scan)"
        rm $PENV_LOCK_FILE
    fi
}

function penv-ify () {
    VIRTUAL_ENV_DIRECTORY_PATH=$1

    mkdir -p .penv/.plugins

    if [ -d "$VIRTUAL_ENV_DIRECTORY_PATH" ]; then
        VIRTUAL_ENV_DIRECTORY_NAME="`basename $VIRTUAL_ENV_DIRECTORY_PATH`"
        echo "$VIRTUAL_ENV_DIRECTORY_NAME" > .penv/default
        builtin cd .penv                                                     && \
            ln -s ../$VIRTUAL_ENV_DIRECTORY_PATH $VIRTUAL_ENV_DIRECTORY_NAME && \
            builtin cd ..
    else
        penv venv-new
    fi

    cd .
}
"""


def execute(args, script=None, env=None):
    # TODO: is cleanup needed before calling exec? (open files, ...)
    command = script or args[0]
    os.execvpe(command, args, env or os.environ)


def command_is_available(command, shell=False):
    try:
        subprocess.check_output(
            command,
            stderr=subprocess.STDOUT,
            shell=shell)
        return True
    except subprocess.CalledProcessError:
        return False
    except Exception:
        return False


@click.group(invoke_without_command=True)
@click.option('--version', '-v', is_flag=True, default=False,
              help=('Prints version'))
@click.option('--startup-script', is_flag=True, default=False,
              help=('Just prints the command to be evaluated by given shell '
                    '(so far only bash supported)'))
@click.pass_context
def cli(ctx, version, startup_script):
    if ctx.invoked_subcommand is None and version:
        return click.echo(VERSION)

    if ctx.invoked_subcommand is None and startup_script:
        return click.echo(startup_script_bash)

    if ctx.invoked_subcommand is None:
        return click.echo(ctx.command.get_help(ctx))

    # Maybe I'll make "place" customizable at some point
    ctx.obj = {
        'place': abspath('.'),
        'startup_script': startup_script,
    }


# $> penv scan
@cli.command('scan')
@click.pass_context
def cli_scan(ctx, env=Penv()):
    penv_exists, place, stream = env.lookup(ctx.obj['place'])
    return click.echo(stream)


# $> penv venv-new
@cli.command('venv-new')
@click.option('--python', '-p', default=None,
              help=('Python executable to be used'))
@click.pass_context
def cli_venv_new(ctx, python, env=Penv()):
    penv_exists, place, stream = env.lookup(ctx.obj['place'])
    if not penv_exists:
        msg = 'Could not find .penv directory (starting from: %s)' % place
        return click.echo(msg)

    python_executable = python or os.environ.get('PYTHON_EXECUTABLE')
    python_command = python_executable or 'python'

    datestamp = datetime.now().strftime('%Y_%m_%d__%H%M%S')
    venv_name = 'venvs/venv_%s' % (datestamp, )
    default_pointer = os.path.join(place, '.penv', 'default')
    with open(default_pointer, 'w') as fd:
        fd.write(venv_name)

    venv_place = os.path.join(place, '.penv', venv_name)
    option = '--prompt="\(%s__%s\)"' % (
        datestamp,
        os.path.basename(place),
    )

    if command_is_available([python_command, '-m', 'venv', '--help']):
        command = [
            python_command, '-m', 'venv', option
        ]
    elif command_is_available(['virtualenv', '--help']):
        command = [
            'virtualenv', venv_place, option
        ] + ([
            '--python=%s' % (python_executable, )
        ] if python_executable else [])
    else:
        raise click.Abort(
            'Cannot find neither of: "%s -m venv" nor "virtualenv"' % (
                python_command,
            )
        )

    return execute(command)



# -*- coding: utf-8 -*-
"""
Definition of plugin registry etc...
"""
import os
import sys
import fnmatch
import importlib

import click

from . import bash


exists = os.path.exists


class Plugin(bash.BashMixin):
    def __init__(self, registry, config):
        self.registry = registry
        self.config = config

    def debug(self, msg, *args):
        msg = str(msg)
        sys.stdout.write('%s\n' % self.bash.echo(msg % args))

    # Invoked on: "penv init"
    #
    # Basically: what todo *after* ".penv" gets activated in *this*
    # directory I'm currently in. Therefore ".penv" is guaranteed to
    # exist when this method is invoked.
    def init(self, root_new):
        return []

    def on_activate(self, root_new, root_old):
        return []

    def on_deactivate(self, root_new, root_old):
        return []


# Very simple register
class PluginRegistry(list):
    def add(self, *plugin_class):
        self.extend(plugin_class)


registry = PluginRegistry()
# end / register


# buildin plugins
class AvailablePlugins(Plugin):
    """
    Shows information about all loaded plugins.
    """

    def format_plugin_info(self, plugin_class):
        module_path = '{p.__module__}.{p.__name__}'.format(p=plugin_class)
        module = importlib.import_module(plugin_class.__module__)
        info = plugin_class.__doc__ or '-- no docs --'
        return (""
            "{module_path}\n"
            "    Found in {module_file_path}\n"
            "    {info}\n"
        ).format(
            module_path=click.style(module_path, fg='green', bold=True),
            module_file_path=module.__file__,
            info=info,
        )

    def function_body(self):
        return self.bash.echo("\n".join(
            self.format_plugin_info(pcls) for pcls in self.registry
        ))

    def on_activate(self, root_new, root_old):
        return [
            self.bash.function("penv-plugins", self.function_body()),
        ]

    def on_deactivate(self, root_new, root_old):
        return [
            self.bash.unset_f("penv-plugins"),
        ]
# end / buildin plugins


# LOADING PLUGINS
def load(plugin_places, config):

    def filter_python_files(file_name):
        return fnmatch.fnmatch(file_name, "*.py")

    def load_builtins():
        registry.add(AvailablePlugins)

    def load_plugins_from(path):
        plugin_files = os.listdir(path)
        plugin_files = filter(filter_python_files, plugin_files)

        for plugin_file in plugin_files:
            module_name = plugin_file.rstrip(".py")
            importlib.import_module(module_name)

    def load():
        for place in plugin_places:
            if exists(place):
                sys.path.insert(0, place)
                load_plugins_from(place)
                sys.path.pop(0)

    if not registry:
        load_builtins()
        load()

    registry_copy = tuple(registry)
    return [pclass(registry_copy, config) for pclass in registry]
# end / LOADING PLUGINS
